"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createWidgets = exports.retryDataBinds = exports.findWidgets = exports.getDom = exports.checkVisible = exports.WidgetsContext = exports.WIDGET_API_MAP = exports.widgetKeys = void 0;
var tslib_1 = require("tslib");
var React = tslib_1.__importStar(require("react"));
var mobx_1 = require("mobx");
var lodash_es_1 = require("lodash-es");
var eventProxy_1 = require("./eventProxy");
var adapter_1 = require("./adapter");
var common_1 = require("./common");
var WIDGET_API_LIST = [
    'getConfig',
    'findWidgets',
    'getWidgetsByType',
    'extends',
    'addWidgetsListener',
    'removeWidgetsListener',
    'getOwnerWidget',
    'getDom',
    'domRef',
    '_listeners',
    '_getInstanceRef',
    '_scope',
];
exports.widgetKeys = tslib_1.__spreadArray(tslib_1.__spreadArray([], tslib_1.__read(WIDGET_API_LIST), false), ['_key', 'parent', '_visible', '_parentId', '_waForKey'], false);
exports.WIDGET_API_MAP = WIDGET_API_LIST.reduce(function (map, key) {
    map[key] = true;
    return map;
}, {});
exports.WidgetsContext = React.createContext({ parent: null });
function checkVisible(_a) {
    var id = _a.id, _b = _a._visible, _visible = _b === void 0 ? undefined : _b;
    return _visible !== false && _visible !== '';
}
exports.checkVisible = checkVisible;
function getDom(element, options) {
    if (!element) {
        console.warn('getDom 接口未传入有效的 element');
        return Promise.resolve({});
    }
    var result = {};
    if (options.id)
        result.id = element.id;
    if (options.dataset)
        result.dataset = element.dataset;
    if (options.rect) {
        var rect = element.getBoundingClientRect();
        Object.assign(result, {
            left: rect.left,
            right: rect.right,
            top: rect.top,
            bottom: rect.bottom,
        });
    }
    if (options.size) {
        var rect = element.getBoundingClientRect();
        Object.assign(result, {
            width: rect.width,
            height: rect.height,
        });
    }
    if (options.scrollOffset) {
        Object.assign(result, {
            scrollLeft: element.scrollLeft,
            scrollTop: element.scrollTop,
        });
    }
    if (options.properties && Array.isArray(options.properties)) {
        options.properties.forEach(function (propName) {
            result[propName] = element.getAttribute(propName);
        });
    }
    if (options.computedStyle && Array.isArray(options.computedStyle)) {
        var computedStyle_1 = adapter_1.runtime.getComputedStyle(element);
        options.computedStyle.forEach(function (propName) {
            result[propName] = computedStyle_1[propName];
        });
    }
    return Promise.resolve(result);
}
exports.getDom = getDom;
function findWidgets(widget, parentType, filterFn, includeInvisibleDescendants) {
    if (!widget)
        return [];
    var _a = widget.children, children = _a === void 0 ? [] : _a;
    var matched = children.filter(filterFn);
    // 过滤掉不可见后代
    if (!includeInvisibleDescendants) {
        children = children.filter(function (item) { return checkVisible(item); });
    }
    children.forEach(function (child) {
        // 如果递归过程中发现了自己，则停止递归
        matched = matched.concat(findWidgets(child, parentType, filterFn, includeInvisibleDescendants));
    });
    // 过滤掉不可见的匹配项
    if (!includeInvisibleDescendants) {
        matched = matched.filter(function (item) { return checkVisible(item); });
    }
    return matched;
}
exports.findWidgets = findWidgets;
var retryQueue = [];
// 递归执行
function retryDataBinds(tryTime) {
    if (tryTime === void 0) { tryTime = 10; }
    var fn = retryQueue.shift();
    if (!fn || tryTime <= 0) {
        retryQueue = [];
        return;
    }
    try {
        fn({ silent: tryTime > 1 });
    }
    catch (e) {
        console.error('retryDataBinds', e);
    }
    retryDataBinds(tryTime - 1);
}
exports.retryDataBinds = retryDataBinds;
function createWidgets(ctx, widgetProps, dataBinds, context) {
    var _a = createWidgetTree(ctx.nodeMap, ctx.widgetProps, widgetProps, dataBinds), nodeTree = _a.nodeTree, nodeMap = _a.nodeMap;
    var widgets = runFor(ctx.widgets, nodeTree, {}, null, null, context);
    return { widgets: widgets, nodeMap: nodeMap };
    /**
     * @param {*} pre widgets
     * @param {*} curForNode
     * @param {*} forItems
     * @param {*} parentLevelWidgets
     * @param {*} parentWidget
     * @returns top level widgets or for dispose
     */
    function runFor(preWidgets, curForNode, forItems, parentLevelWidgets, parentWidget, dataContext) {
        var nodeId = curForNode.id;
        var ctx = {
            preWidgets: preWidgets,
            parentLevelWidgets: parentLevelWidgets,
            parentWidget: parentWidget,
        };
        if (!curForNode.value) {
            return createSubTree(ctx, curForNode, {}, dataContext);
        }
        var _waForKey = curForNode.value._waForKey;
        var timer = {};
        var dispose = (0, mobx_1.autorun)(function (r) {
            var forList = [];
            try {
                clearTimeout(timer[nodeId]);
                /**
                 * untrace 收集context 阶段不添加observer
                 * 消费数据阶段再添加
                 */
                var dataContext_1 = (0, mobx_1.untracked)(function () { return (0, common_1.generateDataContext)(parentWidget); });
                var list = dataBinds[nodeId]._waFor(forItems, undefined, dataContext_1);
                if (!Array.isArray(list)) {
                    timer[nodeId] = setTimeout(function () {
                        console.warn(nodeId, 'For 循环绑定的数据并不是数组，请检查');
                    }, 1000);
                }
                else {
                    forList = list;
                }
            }
            catch (e) {
                if ((window === null || window === void 0 ? void 0 : window.__WEDA_LOG_LEVEL__) === 'debug') {
                    timer[nodeId] = setTimeout(function () {
                        console.debug("".concat(nodeId, "\u5FAA\u73AF\u7ED1\u5B9A\u9519\u8BEF"), e);
                    }, 1000);
                }
            }
            // r.trace();
            // 让 forList 进行监听
            forList.forEach(function () { });
            (0, mobx_1.untracked)(function () {
                // dispose widgets before reused instead
                // disposeWidgets(parentLevelWidgets[curForNode.id]);
                var exsitMap = forList.reduce(function (map, item) {
                    if (item === null || item === void 0 ? void 0 : item[_waForKey]) {
                        map[item[_waForKey]] = true;
                    }
                    return map;
                }, {});
                var forWidgets = (parentLevelWidgets === null || parentLevelWidgets === void 0 ? void 0 : parentLevelWidgets[nodeId]) || [];
                var existedWidgetIndex = [];
                var existedWidgets = [];
                var extraWidgetsIndex = [];
                var extraWidgets = [];
                forWidgets.forEach(function (widget, index) {
                    if (exsitMap[widget._key]) {
                        existedWidgetIndex.push(index);
                        existedWidgets.push(widget);
                        // need to use uqique key
                        exsitMap[widget._key] = undefined;
                    }
                    else {
                        extraWidgetsIndex.push(index);
                        extraWidgets.push(widget);
                    }
                });
                var extraWidgetsIndexMap = extraWidgetsIndex.reduce(function (map, item) {
                    map[item] = true;
                    extraWidgets.push(forWidgets[item]);
                    return map;
                }, {});
                // clean nodes of previouse for run
                dfsTree(curForNode, function (node) {
                    var arr = parentLevelWidgets[node.id];
                    // arr.splice(0, arr.length);
                    (0, lodash_es_1.remove)(arr, function (_, index) {
                        return extraWidgetsIndexMap[index];
                    });
                }, undefined);
                // 清理根 for 的 autorun, 并递归清理子节点
                extraWidgets.map(function (w) {
                    disposeWidgets([w]);
                    parentWidget && (0, lodash_es_1.remove)(parentWidget.children, function (_a) {
                        var id = _a.id;
                        return id === nodeId;
                    });
                });
                forList.forEach(function (item, index) {
                    var _a;
                    var subForItems = tslib_1.__assign(tslib_1.__assign({}, forItems), (_a = {}, _a[nodeId] = item, _a.forIndexes = tslib_1.__spreadArray([], tslib_1.__read(((forItems === null || forItems === void 0 ? void 0 : forItems['forIndexes']) || [])), false).concat(index), _a));
                    createSubTree(ctx, curForNode, subForItems, dataContext);
                });
                // 非初始化时遇到需要重新构建 dataBinds
                retryDataBinds();
            });
        });
        return dispose;
    }
    function createSubTree(ctx, curForNode, subForItems, dataContext) {
        var preWidgets = ctx.preWidgets, parentLevelWidgets = ctx.parentLevelWidgets, parentWidget = ctx.parentWidget;
        var widgets = { _disposers: [] };
        // traverse down the tree to set all widgets
        dfsTree(curForNode, function (node, parentNode, cache) {
            var e_1, _a;
            var _b, _c;
            var parentForWidgetArr = (parentLevelWidgets === null || parentLevelWidgets === void 0 ? void 0 : parentLevelWidgets[node.id]) || [];
            var _waForKey = node.value._waForKey;
            var key = (_b = subForItems[node.id]) === null || _b === void 0 ? void 0 : _b[_waForKey];
            var index = cache[parentNode === null || parentNode === void 0 ? void 0 : parentNode.id]
                ? cache[parentNode.id].index
                : parentForWidgetArr.findIndex(function (widget) { return key && widget._key === key; });
            var forExsitWidget = index !== -1 ? parentForWidgetArr[index] : null;
            var preExistWidget = node.exist
                ? Array.isArray(preWidgets === null || preWidgets === void 0 ? void 0 : preWidgets[node.id])
                    ? (0, lodash_es_1.get)(preWidgets === null || preWidgets === void 0 ? void 0 : preWidgets[node.id], subForItems.forIndexes)
                    : preWidgets === null || preWidgets === void 0 ? void 0 : preWidgets[node.id]
                : null;
            var existedWidget = forExsitWidget || preExistWidget || null;
            if (forExsitWidget) {
                cache[node.id] = {
                    index: index,
                };
            }
            if (node.forCount === curForNode.forCount) {
                // Leaf node
                var w_1 = existedWidget;
                if (!existedWidget) {
                    w_1 = (0, mobx_1.observable)(tslib_1.__assign(tslib_1.__assign({}, node.value), { children: [], id: node.id, _key: key, parent: null }));
                    if (node === curForNode) {
                        w_1._disposers = [];
                    }
                    // 更新 scope
                    w_1._scope = {
                        id: node.id,
                        dataContext: dataContext[node.id] || undefined,
                    };
                    w_1._listeners = new eventProxy_1.EventProxy();
                    // 提供一个给 Node 挂载 API 的方式
                    (0, mobx_1.untracked)(function () {
                        if (!w_1.getConfig) {
                            // 兼容非复合组件调用 getConfig 能力
                            w_1.getConfig = function () { return ({}); };
                        }
                        w_1.findWidgets = function (type, includeInvisibleDescendants) {
                            return findWidgets(w_1, w_1.widgetType, type, includeInvisibleDescendants);
                        };
                        w_1.getWidgetsByType = function (type, includeInvisibleDescendants) {
                            return w_1.findWidgets(function (currentWidget) { return currentWidget.widgetType === type; }, includeInvisibleDescendants);
                        };
                        w_1.extends = function (name, fnOrData) {
                            return Object.defineProperty(w_1, name, {
                                value: fnOrData,
                                writable: true,
                            });
                        };
                        w_1.addWidgetsListener = function (name, listener) { return w_1._listeners.on(name, listener); };
                        w_1.removeWidgetsListener = function (name, listener) { return w_1._listeners.cancel(name, listener); };
                        w_1.getOwnerWidget = function () { return null; }; // 默认带上一个 widget
                        // eslint-disable-next-line react/display-name
                        w_1._getInstanceRef = function () { return null; }; // 默认初始值
                        Object.defineProperty(w_1, '_methods', {
                            get: function () {
                                var instance = this._getInstanceRef();
                                return instance === null || instance === void 0 ? void 0 : instance.current.methods;
                            },
                        });
                    });
                    var parent_1 = parentNode ? widgets[parentNode.id] : parentWidget;
                    if (parent_1) {
                        w_1.parent = parent_1;
                        // 只有可显示 visible 的才存入 children 里
                        if (checkVisible(w_1)) {
                            parent_1.children.push(w_1);
                        }
                    }
                }
                else {
                    if (!forExsitWidget && preExistWidget) {
                        if (node.changed) {
                            try {
                                for (var _d = tslib_1.__values(node.changed), _e = _d.next(); !_e.done; _e = _d.next()) {
                                    var key_1 = _e.value;
                                    w_1[key_1] = node.value[key_1];
                                }
                            }
                            catch (e_1_1) { e_1 = { error: e_1_1 }; }
                            finally {
                                try {
                                    if (_e && !_e.done && (_a = _d.return)) _a.call(_d);
                                }
                                finally { if (e_1) throw e_1.error; }
                            }
                        }
                        // 仅仅为了受控模式下更新 dataContext
                        if (dataContext === null || dataContext === void 0 ? void 0 : dataContext[w_1.id]) {
                            w_1._scope.dataContext = dataContext[w_1.id];
                        }
                    }
                    disposeWidgets([existedWidget], true);
                }
                if (!forExsitWidget) {
                    (_c = parentForWidgetArr.push) === null || _c === void 0 ? void 0 : _c.call(parentForWidgetArr, w_1);
                }
                widgets[node.id] = w_1;
                // Setup data binds
                Object.keys(dataBinds[node.id] || {})
                    .sort(function (a, b) {
                    return a.length - b.length > 0 ? 1 : -1;
                })
                    .map(function (prop) {
                    if (prop === '_waFor') {
                        return;
                    }
                    var timer = null;
                    function getBindData(curForNodeId, options) {
                        if (options === void 0) { options = { silent: false }; }
                        var disposeError = false;
                        var dispose = (0, mobx_1.autorun)(function (reaction) {
                            try {
                                clearTimeout(timer);
                                /**
                                 * untrace 收集context 阶段不添加observer
                                 * 消费数据阶段再添加
                                 */
                                var dataContext_2 = (0, mobx_1.untracked)(function () { return (0, common_1.generateDataContext)(w_1); });
                                // Computed data bind in the next tick since data bind may read widgets data
                                var value_1 = dataBinds[node.id][prop](subForItems, undefined, dataContext_2);
                                var paths_1 = prop.split('.').filter(function (key) { return !!key; });
                                if (paths_1.length > 1) {
                                    // 一定要 untracked 不然爆栈了
                                    (0, mobx_1.untracked)(function () {
                                        var pre = (0, lodash_es_1.get)(w_1, prop);
                                        (0, lodash_es_1.set)(w_1, prop, value_1);
                                        /**
                                         * 深层的对象内部值变化不会触发 obverver，因此浅拷贝对象
                                         */
                                        if (Array.isArray(w_1[paths_1[0]])) {
                                            if (!(0, lodash_es_1.isEqual)(pre, value_1)) {
                                                w_1[paths_1[0]] = tslib_1.__spreadArray([], tslib_1.__read(w_1[paths_1[0]]), false);
                                            }
                                        }
                                    });
                                }
                                else {
                                    // 普通 key 直接赋值
                                    w_1[prop] = value_1;
                                }
                                disposeError = false;
                            }
                            catch (e) {
                                if (prop === '_visible') {
                                    w_1[prop] = false;
                                }
                                else {
                                    retryQueue.push(function (opts) {
                                        reaction.dispose();
                                        return getBindData(curForNodeId, opts);
                                    });
                                    disposeError = true;
                                }
                                if (!(options === null || options === void 0 ? void 0 : options.silent)) {
                                    if ((window === null || window === void 0 ? void 0 : window.__WEDA_LOG_LEVEL__) === 'debug') {
                                        timer = setTimeout(function () {
                                            console.debug("\u6570\u636E\u7ED1\u5B9A\u8BA1\u7B97\u9519\u8BEF ".concat(w_1.id, ".").concat(prop), e);
                                            timer = null;
                                        }, 1000);
                                    }
                                }
                            }
                        });
                        if (!disposeError) {
                            if (curForNodeId) {
                                widgets[curForNode.id]._disposers.push(dispose);
                            }
                            else {
                                widgets._disposers.push(dispose);
                            }
                        }
                    }
                    getBindData(curForNode.id);
                });
            }
            else if (!existedWidget) {
                if (parentLevelWidgets) {
                    var len = parentLevelWidgets[node.id].push([]);
                    widgets[node.id] = parentLevelWidgets[node.id][len - 1];
                }
                else {
                    widgets[node.id] = (0, mobx_1.observable)([]);
                }
            }
            else {
                widgets[node.id] = existedWidget;
            }
        }, undefined);
        // run for of next level
        dfsTree(curForNode, function (node, parentNode) {
            if (node.forCount === curForNode.forCount + 1 && dataBinds[node.id] && dataBinds[node.id]._waFor) {
                widgets[node.id]._disposers = { dataBinds: [] };
                var dispose = runFor(preWidgets, node, subForItems, widgets, node.parent && widgets[node.parent.id], dataContext);
                curForNode.id && widgets[curForNode.id]._disposers.push(dispose);
            }
        }, undefined);
        return widgets;
    }
}
exports.createWidgets = createWidgets;
function createWidgetTree(preNodes, preWidgets, widgets, dataBinds) {
    var virtualRoot = { children: [], forCount: 0 };
    var nodes = Object.keys(widgets).reduce(function (result, id) {
        var currentValue = widgets[id];
        result[id] = {
            id: id,
            value: currentValue,
            children: [],
            parent: null,
            forCount: 0,
            exist: false,
            changed: [],
        };
        return result;
    }, {});
    // Create widgets tree API
    Object.keys(nodes).map(function (id) {
        var curNode = nodes[id];
        var parent = nodes[widgets[id]._parentId];
        if (!parent) {
            virtualRoot.children.push(curNode);
            return;
        }
        curNode.parent = parent;
        parent.children.push(curNode);
    });
    virtualRoot.children.map(addForCount);
    // dfs, add forCount
    function addForCount(node) {
        var _a;
        if (node.parent) {
            node.forCount = node.parent.forCount;
        }
        if ((_a = dataBinds[node.id]) === null || _a === void 0 ? void 0 : _a._waFor) {
            node.forCount += 1;
        }
        node.children.map(addForCount);
    }
    dfsTree(virtualRoot, function (node, parent) {
        var currentValue = node.value;
        var preValue = preWidgets === null || preWidgets === void 0 ? void 0 : preWidgets[node.id];
        var preNode = preNodes === null || preNodes === void 0 ? void 0 : preNodes[node.id];
        var exist = false;
        var changed = [];
        // 多级for绑定的情况会发生父for不变化，子for更新的情况，此时子for生成会有异常，因此复用只复用一层保证渲染正常
        // https://lcap.cloud.tencent.com/preview/?u=https%3A%2F%2Fci-source-1303824488.cos.ap-shanghai.myqcloud.com%2Flcp%2Fapp-h057yqaw%2Fappdata-1666253755745.json&appId=app-h057yqaw#/nav
        if ((!parent || (parent === null || parent === void 0 ? void 0 : parent.exist)) && preValue && preNode) {
            if (preValue.widgetType === (currentValue === null || currentValue === void 0 ? void 0 : currentValue.widgetType) &&
                preValue._parentId === currentValue._parentId &&
                preNode.forCount === node.forCount) {
                exist = true;
                for (var key in currentValue) {
                    if (!exports.WIDGET_API_MAP[key] && !(0, lodash_es_1.isEqual)(preValue[key], currentValue[key])) {
                        changed.push(key);
                    }
                }
            }
            node.exist = exist;
            node.changed = changed;
        }
    }, undefined);
    return { nodeTree: virtualRoot, nodeMap: nodes };
}
function dfsTree(node, fn, parent, cache, scope) {
    if (cache === void 0) { cache = {}; }
    if (scope === void 0) { scope = { dataContext: {} }; }
    if (node.value) {
        fn(node, parent, cache, scope);
    }
    node.children.map(function (e) { return dfsTree(e, fn, node.value ? node : null, cache, scope); });
}
/**
 * dispose autorun
 * TODO: 是否要同步清理 parent children 引用，防止内存泄漏
 */
function disposeWidgets(widgets, noRecursive) {
    if (widgets === void 0) { widgets = []; }
    if (noRecursive === void 0) { noRecursive = false; }
    widgets.forEach(function (widget) {
        /**
         * 解除 parent 依赖
         */
        // widget.parent = null;
        var disposers = widget._disposers;
        if (disposers) {
            disposers.map(function (dispose) { return dispose(); });
            disposers.splice(0, disposers.length);
        }
        if (!noRecursive) {
            disposeWidgets(widget.children);
            /**
             * 解除children 的引用
             */
            // widget.children = [];
        }
    });
}
//# sourceMappingURL=widgets.js.map